"""Module in charge of generating mocks."""

import os
import re
from copy import deepcopy

from jinja2 import Environment
from jinja2 import PackageLoader
from pycparser import c_ast as node
from pycparser.c_generator import CGenerator

from .inspect import PRIMITIVE_TYPES
from . import __version__


SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

HEADER_FILE = 'nala_mocks.h'
SOURCE_FILE = 'nala_mocks.c'
LINKER_FILE = 'nala_mocks.ldflags'


def is_ellipsis(param):
    return isinstance(param, node.EllipsisParam)


def decl(name, type):
    return node.Decl(name, [], [], [], type, None, None)


def function_ptr_decl(name, return_type, parameters):
    return decl(
        name, node.PtrDecl([], node.FuncDecl(node.ParamList(parameters), return_type))
    )


def bool_param(name):
    return decl(name, node.TypeDecl(name, [], node.IdentifierType(['bool'])))


def set_member(name):
    return decl(name,
                node.TypeDecl(name,
                              [],
                              node.IdentifierType(['struct nala_set_param'])))


def in_assert_member(name, param_actual, param_expected):
    name = f'{name}_in_assert'

    return function_ptr_decl(
        name,
        void_type(name),
        [
            param_actual,
            param_expected,
            decl('size', node.TypeDecl('size', [], node.IdentifierType(['size_t'])))
        ])


def out_copy_member(name, param_dst, param_src):
    name = f'{name}_out_copy'

    return function_ptr_decl(
        name,
        void_type(name),
        [
            param_dst,
            param_src,
            decl('size', node.TypeDecl('size', [], node.IdentifierType(['size_t'])))
        ])


def va_list_param(name):
    return decl(None,
                node.TypeDecl(name, [], node.IdentifierType(['va_list'])))


def is_variadic_func(params):
    if params:
        return is_ellipsis(params[-1])
    else:
        return False


def void_type(name):
    return node.TypeDecl(name, [], node.IdentifierType(['void']))


def int_type(name):
    return node.TypeDecl(name, [], node.IdentifierType(['int']))


def rename_return_type(return_type, name):
    return_type = deepcopy(return_type)
    type_decl = return_type

    while not isinstance(type_decl, node.TypeDecl):
        type_decl = type_decl.type

    type_decl.declname = name

    return return_type


def create_implementation_params(params):
    return [
        va_list_param('__nala_va_list')
        if is_ellipsis(param)
        else param
        for param in params
    ]


def get_guard_name(filename):
    slug = re.sub(r'[^a-zA-Z0-9]', '_', os.path.normpath(os.path.relpath(filename)))

    return re.sub(r'_+', '_', slug).upper().strip('_')


def generate_includes(includes, directory):
    include_lines = []

    for path, system in includes:
        if system:
            include_lines.append(f'#include <{path}>')
        else:
            include_lines.append(f'#include "{os.path.relpath(path, directory)}"')

    return '\n'.join(include_lines)


def read_nala_c():
    path = os.path.join(SCRIPT_DIR, 'templates', 'nala.c')

    with open(path, 'r') as fin:
        return fin.read()


def load_typedefs(ast):
    typedefs = {}

    for item in ast:
        if isinstance(item, node.Typedef):
            typedefs[item.name] = item

    return typedefs


class StructAssert:

    def __init__(self, struct):
        self.name = struct[0]
        self.assert_eq_members = []
        self.assert_array_eq_members = []
        self.assert_struct_members = []
        self.assert_struct_typedef_members = []

        for member in struct[1]:
            if member[0] == 'assert-eq':
                self.assert_eq_members.append(member[1])
            elif member[0] == 'assert-array-eq':
                self.assert_array_eq_members.append(member[1])
            elif member[0] == 'assert-struct':
                self.assert_struct_members.append(member[1:])
            elif member[0] == 'assert-struct-typedef':
                self.assert_struct_typedef_members.append(member[1:])


class StructTypedefAssert:

    def __init__(self, typedef):
        self.name = typedef[0]
        self.assert_eq_members = []
        self.assert_array_eq_members = []
        self.assert_struct_members = []
        self.assert_struct_typedef_members = []

        for member in typedef[1]:
            if member[0] == 'assert-eq':
                self.assert_eq_members.append(member[1])
            elif member[0] == 'assert-array-eq':
                self.assert_array_eq_members.append(member[1])
            elif member[0] == 'assert-struct':
                self.assert_struct_members.append(member[1:])
            elif member[0] == 'assert-struct-typedef':
                self.assert_struct_typedef_members.append(member[1:])


class FunctionMock:

    DECL_MARKER = "// NALA_DECLARATION"
    IMPL_MARKER = "// NALA_IMPLEMENTATION"

    def __init__(self,
                 function,
                 has_implementation,
                 real_variadic_function,
                 struct_names,
                 struct_typedef_names):
        self.function = function
        self.func_name = function.name
        self.real_variadic_function = real_variadic_function
        self.struct_names = struct_names
        self.struct_typedef_names = struct_typedef_names

        self.wrapped_func = f'__wrap_{self.func_name}'
        self.real_func = f'__real_{self.func_name}'

        self.state_name = f'nala_mock_{self.func_name}'

        self.typedefs = load_typedefs(function.file_ast)
        self.func_decl = self.function.declaration.type
        self.func_params = self.func_decl.args.params if self.func_decl.args else []
        self.assign_names_to_unnamed_params(self.func_params)
        self.is_variadic_func = is_variadic_func(self.func_params)

        if has_implementation is not None:
            self.has_implementation = has_implementation
        elif self.is_variadic_func:
            if real_variadic_function:
                self.has_implementation = True
            else:
                self.has_implementation = False
        else:
            self.has_implementation = True

        self.params_struct = [
            decl(param.name, node.PtrDecl([], param.type.type))
            if isinstance(param.type, node.ArrayDecl)
            else param
            for param in self.func_params
            if not is_ellipsis(param) and param.name
        ]
        self.forward_args = ', '.join(param.name for param in self.params_struct)

        if self.is_variadic_func:
            self.params_struct.append(decl(
                None,
                node.PtrDecl([],
                             node.TypeDecl('vafmt_p',
                                           ['const'],
                                           node.IdentifierType(['char'])))))
            self.forward_args += ', nala_vl'

        # -Wpedantic warns on empty structs.
        if not self.params_struct:
            self.params_struct = [
                decl('dummy',
                     node.TypeDecl('dummy', [], node.IdentifierType(['int'])))
            ]

        return_type = self.func_decl.type
        self.return_value = (
            None
            if isinstance(return_type, node.TypeDecl)
            and isinstance(return_type.type, node.IdentifierType)
            and return_type.type.names[0] == 'void'
            else 'return_value')

        if self.is_variadic_func:
            self.va_list_start_arg_name = self.func_params[-2].name
        else:
            self.va_list_start_arg_name = None

        self.return_value_decl = decl(
            self.return_value,
            rename_return_type(return_type, self.return_value))
        mock_params = self.create_mock_params()
        self.implementation_decl = function_ptr_decl(
            'implementation',
            rename_return_type(return_type, 'implementation'),
            create_implementation_params(self.func_params))
        self.mock_func = self.void_function_decl(f'{self.func_name}_mock',
                                                 mock_params)
        self.mock_once_func = self.int_function_decl(
            f'{self.func_name}_mock_once',
            mock_params)
        self.set_errno_func = self.void_function_decl(
            f'{self.func_name}_mock_set_errno',
            [decl(
                'errno_value',
                node.TypeDecl('errno_value', [], node.IdentifierType(['int'])),
            )])
        self.callback_decl = function_ptr_decl(
            'callback',
            void_type('callback'),
            create_implementation_params(self.func_params))
        self.variadic_func_real_wrapper_decl = node.FuncDecl(
            node.ParamList(create_implementation_params(self.func_params)),
            node.TypeDecl(
                f'{self.func_name}_mock_va_arg_real',
                [],
                return_type))
        self.default_variadic_func_real_wrapper_decl = node.FuncDecl(
            node.ParamList(create_implementation_params(self.func_params)),
            node.TypeDecl(
                f'nala_v{self.func_name}',
                [],
                return_type))
        self.real_decl = self.rename_function(self.real_func)
        self.wrapped_decl = self.rename_function(self.wrapped_func)
        self.instance_members = []
        self.set_params = []
        self.char_pointer_params = []
        self.pointer_params = []
        self.non_pointer_params = []
        self.ignore_params = []

        for param in self.func_params:
            if is_ellipsis(param):
                continue
            elif self.is_struct(param):
                continue
            elif self.is_union(param):
                continue
            elif self.is_va_list(param):
                continue

            if not param.name:
                continue

            self.instance_members.append(bool_param(f'ignore_{param.name}_in'))

            if self.is_char_pointer(param):
                self.char_pointer_params.append(param)
            elif self.is_pointer(param):
                self.pointer_params.append(param)
            else:
                self.non_pointer_params.append(param)

            if self.is_char_pointer_or_non_pointer(param):
                self.ignore_params.append(param.name)

            if not self.is_pointer(param):
                continue

            param_buf = self.rename_param(param, 'buf_p')
            param_actual = self.rename_param(param, 'actual_p')
            param_expected = self.rename_param(param, 'expected_p')
            param_dst = self.rename_param(param, 'dst_p')
            param_src = self.rename_param(param, 'src_p')
            self.instance_members.append(set_member(f'{param.name}_in'))
            self.instance_members.append(in_assert_member(param.name,
                                                          param_actual,
                                                          param_expected))
            self.instance_members.append(set_member(f'{param.name}_out'))
            self.instance_members.append(out_copy_member(param.name,
                                                         param_dst,
                                                         param_src))
            self.set_params.append((param,
                                    param_buf,
                                    param_actual,
                                    param_expected,
                                    param_dst,
                                    param_src,
                                    self.find_check_function(param)))

    def assign_names_to_unnamed_params(self, params):
        for i, param in enumerate(params):
            if not isinstance(param, node.Typename):
                continue

            if self.is_void(param):
                continue

            name = f'nala_{i}'
            param.name = name

            while not isinstance(param, node.TypeDecl):
                param = param.type

            param.declname = name

    def find_check_function(self, param):
        if self.is_pointer_pointer(param):
            return f'nala_mock_assert_pointer'
        elif self.is_char_pointer(param):
            return 'nala_mock_assert_in_string'
        elif self.is_primitive_type_pointer(param):
            return f'nala_mock_assert_{"_".join(param.type.type.type.names)}'
        elif self.is_struct_typedef_pointer(param):
            param = self.lookup_typedef(param.type.type.type.names[0])
            name = f'nala_mock_assert_typedef_struct_{param.name}'

            if name in self.struct_typedef_names:
                return name
        elif self.is_struct_pointer(param):
            name = f'nala_mock_assert_struct_{param.type.type.type.name}'

            if name in self.struct_names:
                return name

        return 'nala_assert_memory'

    def is_primitive_type_pointer(self, param):
        if not isinstance(param.type, node.PtrDecl):
            return False

        try:
            name = ' '.join(param.type.type.type.names)

            if name not in PRIMITIVE_TYPES:
                return False
        except AttributeError:
            return False

        return True

    def is_pointer_pointer(self, param):
        if not isinstance(param.type, node.PtrDecl):
            return False

        if not isinstance(param.type.type, node.PtrDecl):
            return False

        return True

    def void_function_decl(self, name, parameters):
        return node.FuncDecl(node.ParamList(parameters),
                             void_type(name))

    def int_function_decl(self, name, parameters):
        return node.FuncDecl(node.ParamList(parameters),
                             int_type(name))

    def rename_function(self, name):
        return decl(
            name,
            node.FuncDecl(self.func_decl.args,
                          rename_return_type(self.func_decl.type, name)))

    def rename_param(self, param, name):
        param = deepcopy(param)
        param.name = name
        param_next = param

        while not isinstance(param_next, node.TypeDecl):
            param_next = param_next.type

        param_next.declname = name

        return param

    def is_va_list(self, param):
        try:
            if param.type.type.names[0] == '__builtin_va_list':
                return True
        except AttributeError:
            pass

        try:
            name = param.type.type.names[0]

            return self.is_va_list(self.lookup_typedef(name))
        except AttributeError:
            pass

        return False

    def is_char_pointer(self, param):
        if is_ellipsis(param):
            return False
        elif isinstance(param.type, node.PtrDecl):
            if isinstance(param.type.type, (node.FuncDecl, node.PtrDecl)):
                return False
            elif isinstance(param.type.type.type, node.Struct):
                return False
            elif param.type.type.type.names[0] == 'char':
                return True
            else:
                return False
        else:
            return False

    def is_char_pointer_or_non_pointer(self, param):
        if is_ellipsis(param):
            return False
        elif self.is_char_pointer(param):
            return True
        elif self.is_pointer(param):
            return False
        else:
            return True

    def is_void(self, param):
        if is_ellipsis(param):
            return False
        elif self.is_pointer(param):
            return False
        elif self.is_enum(param):
            return False
        elif self.is_struct(param):
            return False
        elif self.is_union(param):
            return False
        else:
            return param.type.type.names[0] == 'void'

    def is_pointer(self, param):
        if isinstance(param.type, (node.PtrDecl, node.ArrayDecl)):
            return True

        try:
            name = param.type.type.names[0]

            return self.is_pointer(self.lookup_typedef(name))
        except AttributeError:
            pass

        return False

    def is_enum(self, param):
        return isinstance(param.type.type, node.Enum)

    def is_struct_pointer(self, param):
        if not self.is_pointer(param):
            return False

        if not self.is_struct(param.type):
            return False

        return True

    def is_struct_typedef_pointer(self, param):
        if not self.is_pointer(param):
            return False

        if not isinstance(param.type.type, node.TypeDecl):
            return False

        if not isinstance(param.type.type.type, node.IdentifierType):
            return False

        param = self.lookup_typedef(param.type.type.type.names[0])

        if param is None:
            return False

        if not isinstance(param.type.type, node.Struct):
            return False

        return True

    def is_struct(self, param):
        try:
            if isinstance(param.type.type, node.Struct):
                return True
        except AttributeError:
            pass

        try:
            name = param.type.type.names[0]

            return self.is_struct(self.lookup_typedef(name))
        except AttributeError:
            pass

        return False

    def is_union(self, param):
        try:
            if isinstance(param.type.type, node.Union):
                return True
        except AttributeError:
            pass

        try:
            name = param.type.type.names[0]

            return self.is_union(self.lookup_typedef(name))
        except AttributeError:
            pass

        return False

    def lookup_typedef(self, name):
        if name in self.typedefs:
            return self.typedefs[name]

    def create_mock_params(self):
        once_params = []
        variable_arguments_params = []

        for param in self.func_params:
            if self.is_void(param):
                continue
            elif self.is_struct(param):
                continue
            elif self.is_union(param):
                continue
            elif self.is_va_list(param):
                continue
            elif self.is_char_pointer_or_non_pointer(param):
                once_params.append(param)
            elif is_ellipsis(param):
                variable_arguments_params.append(decl(
                    None,
                    node.PtrDecl([],
                                 node.TypeDecl('vafmt_p',
                                               ['const'],
                                               node.IdentifierType(['char'])))))
                variable_arguments_params.append(param)

        if not self.is_void(self.return_value_decl):
            once_params.append(self.return_value_decl)

        once_params += variable_arguments_params

        return once_params


class FileGenerator:

    def __init__(self):
        self.code_generator = CGenerator()

        self.jinja_env = Environment(
            loader=PackageLoader('nala', 'templates'),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.jinja_env.filters['render'] = self.code_generator.visit

        self.header_template = self.jinja_env.get_template(f'{HEADER_FILE}.jinja2')
        self.source_template = self.jinja_env.get_template(f'{SOURCE_FILE}.jinja2')

        self.mocks = []
        self.struct_asserts = []
        self.struct_typedef_asserts = []
        self.struct_names = set()
        self.struct_typedef_names = set()
        self.includes = []

    def add_mock(self,
                 mocked_function,
                 has_implementation,
                 real_variadic_function):
        self.mocks.append(FunctionMock(mocked_function,
                                       has_implementation,
                                       real_variadic_function,
                                       self.struct_names,
                                       self.struct_typedef_names))

    def add_include(self, include):
        self.includes.append((include.path, include.system))

    def add_struct(self, struct):
        self.struct_asserts.append(StructAssert(struct))
        self.struct_names.add(
            f'nala_mock_assert_struct_{self.struct_asserts[-1].name}')

    def add_struct_typedef(self, typedef):
        self.struct_typedef_asserts.append(StructTypedefAssert(typedef))
        self.struct_typedef_names.add(
            f'nala_mock_assert_typedef_struct_{self.struct_typedef_asserts[-1].name}')

    def write_to_directory(self, directory):
        os.makedirs(directory, exist_ok=True)

        header_filename = os.path.join(directory, HEADER_FILE)
        source_filename = os.path.join(directory, SOURCE_FILE)
        linker_filename = os.path.join(directory, LINKER_FILE)

        mocks = list(sorted(self.mocks, key=lambda m: m.func_name))
        struct_asserts = list(sorted(self.struct_asserts, key=lambda a: a.name))
        struct_typedef_asserts = list(sorted(self.struct_typedef_asserts,
                                             key=lambda a: a.name))

        header_code = self.header_template.render(
            nala_version=__version__,
            guard_name=get_guard_name(header_filename),
            includes=generate_includes(self.includes, directory),
            mocks=mocks,
            struct_asserts=struct_asserts,
            struct_typedef_asserts=struct_typedef_asserts)

        includes = [
            ('stddef.h', True),
            ('errno.h', True),
            (header_filename, False)
        ]
        source_code = self.source_template.render(
            nala_version=__version__,
            includes=generate_includes(includes, directory),
            nala_c=read_nala_c(),
            mocks=mocks,
            struct_asserts=struct_asserts,
            struct_typedef_asserts=struct_typedef_asserts)

        with open(header_filename, 'w') as fout:
            fout.write(header_code.strip())
            fout.write('\n')

        with open(source_filename, 'w') as fout:
            fout.write(source_code.strip())
            fout.write('\n')

        with open(linker_filename, 'w') as fout:
            fout.write(' '.join([
                f'-Wl,--wrap={mock.function.name}'
                for mock in mocks
            ]))



"""Module in charge of detecting mock usage in expanded source code."""

import os
import sys
import re
from typing import NamedTuple
from typing import Tuple

from pycparser import c_ast as node
from pycparser.c_parser import CParser


class IncludeDirective(NamedTuple):
    path: str
    system: bool

    @classmethod
    def from_source_context(cls, source_context):
        if not source_context:
            return None

        for filename in source_context:
            if not os.path.isabs(filename):
                continue

            dirname, basename = os.path.split(filename)
            fullname = basename

            while True:
                dirname, basename = os.path.split(dirname)

                if 'include' in basename.lower():
                    return cls(fullname, True)

                if not basename:
                    break

                fullname = os.path.join(basename, fullname)

        return cls(os.path.abspath(source_context[-1]), False)

    def __str__(self):
        return '#include ' + (f'<{self.path}>' if self.system else f'"{self.path}"')


class MockedFunction(NamedTuple):
    name: str
    declaration: node.FuncDecl
    file_ast: node.FileAST


class Token(NamedTuple):
    type: str
    value: str
    span: Tuple[int, int]

    def is_punctuation(self, *symbols):
        return self.type == 'PUNCTUATION' and self.value in symbols

    def is_keyword(self, *keywords):
        return self.type == 'KEYWORD' and self.value in keywords

    @property
    def is_prefix(self):
        return self.type == 'KEYWORD' and self.value in (
            'typedef',
            'extern',
            'static',
            'auto',
            'register',
            '__extension__'
        )


def load_param_names(filename):
    param_names = {}

    with open(filename, 'r') as fin:
        for line in fin:
            line = line.strip()

            if not line:
                continue

            if line.startswith('#'):
                continue

            mo = re.match(r'([^(]+)\(([^)]+)\)', line)

            if not mo:
                raise Exception(f"Invalid parameter name line '{line}'.")

            param_names[mo.group(1)] = [
                param_name.strip()
                for param_name in mo.group(2).split(',')
            ]

    return param_names


def rename_parameters(function_declaration, param_names):
    if function_declaration.type.args is None:
        return function_declaration

    for i, param in enumerate(function_declaration.type.args.params):
        if isinstance(param, node.EllipsisParam):
            continue

        param_type = param.type

        if (not param.name
            and isinstance(param_type.type, node.IdentifierType)
            and param_type.type.names == ['void']):
            continue

        if param_names is not None:
            param.name = param_names[i]
        elif param.name is not None:
            if len(param.name) > 2 and param.name.startswith('__'):
                param.name = param.name[2:]

        while not isinstance(param_type, node.TypeDecl):
            param_type = param_type.type

        param_type.declname = param.name

    return function_declaration


PRIMITIVE_TYPES = [
    'char',
    'signed char',
    'unsigned char',
    'short',
    'short int',
    'signed short',
    'signed short int',
    'unsigned short',
    'unsigned short int',
    'int',
    'signed',
    'signed int',
    'unsigned',
    'unsigned int',
    'long',
    'long int',
    'signed long',
    'signed long int',
    'unsigned long',
    'unsigned long int',
    'long long',
    'long long int',
    'signed long long',
    'signed long long int',
    'unsigned long long',
    'unsigned long long int',
    'float',
    'double',
    'long double'
]


class ForgivingDeclarationParser:
    linemarker = re.compile(r'^# \d+ "((?:\\.|[^\\"])*)"((?: [1234])*)$')

    tokens = {
        'LINEMARKER': r"^#.*$",
        'KEYWORD': (
            "\\b(?:auto|break|case|char|const|continue|default|do|double|else|enum|"
            "extern|float"
            "|for|goto|if|int|long|register|return|short|signed|sizeof|static|struct"
            "|switch|typedef|union|unsigned|void|volatile|while|__extension__|"
            "__attribute__|__restrict|__signed__)\\b"
        ),
        'IDENTIFIER': r"\b[a-zA-Z_](?:[a-zA-Z_0-9])*\b",
        'CHARACTER': r"L?'(?:\\.|[^\\'])+'",
        'STRING': r'L?"(\\"|\\\\|.)*?"',
        'INTEGER': r"(?:0[xX][a-fA-F0-9]+|[0-9]+)[uUlL]*",
        'FLOAT': (
            r"(?:[0-9]+[Ee][+-]?[0-9]+|[0-9]*\.[0-9]+(?:[Ee][+-]?[0-9]+)?|[0-9]+\."
            r"[0-9]*(?:[Ee][+-]?[0-9]+)?)[fFlL]?"
        ),
        'PUNCTUATION': (
            r"\.\.\.|>>=|<<=|\+=|-=|\*=|/=|%=|&=|\^=|\|=|>>|<<|\+\+|--|->|&&|\|\|""|<="
            r"|>=|==|!=|;|\{|\}|,|:|=|\(|\)|\[|\]|\.|&|!|~|-|\+|\*|/|%|<|>|\^|\||\?"
        ),
        'SPACE': r"[ \t\v\n\f]*",
        'IGNORE': r".+?",
    }

    ignored_tokens = 'SPACE', 'IGNORE'

    regex = re.compile(
        '|'.join(f'(?P<{token}>{pattern})' for token, pattern in tokens.items()),
        flags=re.MULTILINE)

    def __init__(self, source_code, functions, rename_parameters_file=None):
        self.source_code = source_code
        self.functions = functions
        self.token_stream = self.tokenize(source_code)
        self.previous = None
        self.current = None

        self.chunks_to_erase = []
        self.bracket_stack = []
        self.source_context = []
        self.typedefs_code = ['typedef int __builtin_va_list;']
        self.typedefs = {}
        self.structs_code = []
        self.structs = []
        self.struct_typedefs = []
        self.includes = []
        self.filename = None

        self.cparser = CParser()
        self.param_names = None

        if rename_parameters_file is not None:
            self.param_names = load_param_names(rename_parameters_file)

        self.func_names = []
        self.func_signatures = []
        self.func_source_contexts = []
        self.file_ast = None
        self.mocked_functions = []
        self.parse()

        if self.functions:
            for function in sorted(functions):
                print(f"error: Mocked function '{function}' undeclared. Missing include?",
                      file=sys.stderr)

            raise Exception(
                'Unable to find declarations of all mocked functions. Add missing '
                'includes to the test file.')

    @classmethod
    def tokenize(cls, source_code):
        for match in cls.regex.finditer(source_code):
            if match.lastgroup not in cls.ignored_tokens:
                yield Token(match.lastgroup, match.group().strip(), match.span())

    def parse(self):
        while self.next():
            if self.current.is_keyword('typedef'):
                self.parse_typedef()

            parsed = self.parse_function_declaration_or_struct()

            if parsed is not None:
                self.func_names.append(parsed[0])
                self.func_signatures.append(parsed[1])
                self.func_source_contexts.append(self.source_context[:])
                self.functions.remove(parsed[0])

            if not self.functions:
                break

            while self.bracket_stack or not self.current.is_punctuation(';', '}'):
                self.next()

        if self.functions:
            return

        code = '\n'.join(
            self.typedefs_code + self.structs_code + self.func_signatures)
        self.file_ast = self.cparser.parse(code)
        items = zip(self.func_names, self.func_source_contexts)
        func_offset = len(self.typedefs_code + self.structs_code)

        for i, (func_name, _) in enumerate(items, func_offset):
            if self.param_names is None:
                func_declaration = self.file_ast.ext[i]
            else:
                func_declaration = rename_parameters(
                    self.file_ast.ext[i],
                    self.param_names.get(func_name))

            self.mocked_functions.append(MockedFunction(
                func_name,
                func_declaration,
                self.file_ast))

        self.load_typedefs()
        self.load_structs()

    def is_primitive_type(self, member):
        if member is None:
            return False

        if not isinstance(member.type, node.TypeDecl):
            return False

        if not isinstance(member.type.type, node.IdentifierType):
            return False

        name = ' '.join(member.type.type.names)

        if name in PRIMITIVE_TYPES:
            return True

        return self.is_primitive_type(self.lookup_typedef(name))

    def is_array(self, member):
        if not isinstance(member.type, node.ArrayDecl):
            return False

        if member.type.dim is None:
            return False

        return True

    def is_struct(self, member):
        if not isinstance(member.type, node.TypeDecl):
            return False

        if not isinstance(member.type.type, node.Struct):
            return False

        return True

    def is_struct_typedef(self, member):
        if member is None:
            return False

        if self.is_struct(member):
            return True

        if not isinstance(member.type, node.TypeDecl):
            return False

        if not isinstance(member.type.type, node.IdentifierType):
            return False

        member = self.lookup_typedef(member.type.type.names[0])

        return self.is_struct_typedef(member)

    def lookup_typedef(self, name):
        if name in self.typedefs:
            return self.typedefs[name]

    def load_struct_member(self, member):
        items = []

        if self.is_array(member):
            items.append(['assert-array-eq', member.name])
        elif self.is_primitive_type(member):
            if member.bitsize is None:
                items.append(['assert-eq', member.name])
        elif self.is_struct(member):
            if member.type.type.name is not None:
                items.append(['assert-struct', member.name, member.type.type.name])
            else:
                for item in self.load_struct_members(member.type.type):
                    item[1] = f'{member.name}.{item[1]}'
                    items.append(item)
        elif self.is_struct_typedef(member):
            items.append(
                ['assert-struct-typedef', member.name, member.type.type.names[0]])

        return items

    def load_struct_members(self, struct):
        if not struct.decls:
            return []

        items = []

        for member in struct.decls:
            if member.name is None:
                continue

            items += self.load_struct_member(member)

        return items

    def load_structs(self):
        for item in self.file_ast:
            if isinstance(item, node.Typedef):
                if isinstance(item.type, node.TypeDecl):
                    if isinstance(item.type.type, node.Struct):
                        if item.type.type.decls is not None:
                            items = self.load_struct_members(item.type.type)
                            self.struct_typedefs.append((item.name, items))
            elif isinstance(item, node.Decl):
                if isinstance(item.type, node.Struct):
                    items = self.load_struct_members(item.type)
                    self.structs.append((item.type.name, items))

    def load_typedefs(self):
        for item in self.file_ast:
            if isinstance(item, node.Typedef):
                self.typedefs[item.name] = item

    def next(self):
        self.previous = self.current
        self.current = next(self.token_stream, None)

        if not self.current:
            return None

        if self.current.type == 'PUNCTUATION':
            if self.current.value in '({[':
                self.bracket_stack.append(')}]'['({['.index(self.current.value)])
            elif self.bracket_stack and self.current.value == self.bracket_stack[-1]:
                self.bracket_stack.pop()
        elif self.current.type == 'LINEMARKER':
            filename, flags = self.linemarker.match(self.current.value).groups()

            if self.filename is None:
                self.filename = filename

            if '1' in flags:
                self.source_context.append(filename)

                if len(self.source_context) == 2:
                    if self.source_context[0] == self.filename:
                        if self.source_context[-1] != '<built-in>':
                            self.includes.append(
                                IncludeDirective.from_source_context(
                                    self.source_context))
            elif '2' in flags:
                self.source_context.pop()

            try:
                self.source_context[-1] = filename
            except IndexError:
                self.source_context.append(filename)

            self.mark_for_erase(*self.current.span)
            self.next()
        elif self.current.is_keyword('__attribute__'):
            begin = self.current.span[0]
            stack_depth = len(self.bracket_stack)
            self.next()

            while len(self.bracket_stack) > stack_depth:
                self.next()

            self.mark_for_erase(begin, self.current.span[1])
        elif self.current.is_keyword('__extension__', '__restrict', '__signed__'):
            self.mark_for_erase(*self.current.span)
            self.next()

        return self.current

    def parse_typedef(self):
        begin = self.current.span[0]

        while self.bracket_stack or not self.current.is_punctuation(';'):
            self.next()

        code = self.read_source_code(begin, self.current.span[1])
        self.typedefs_code.append(code)

    def parse_struct(self, begin, _name):
        while self.bracket_stack:
            self.next()

        code = self.read_source_code(begin, self.current.span[1]) + ';'

        if self.is_in_header_file():
            self.structs_code.append(code)

    def is_in_header_file(self):
        return len(self.source_context) > 1

    def parse_function_declaration_or_struct(self):
        while self.current.is_prefix:
            self.next()

        begin = self.current.span[0]
        return_type = []

        if self.current.is_keyword('struct'):
            if self.next() and self.current.type == 'IDENTIFIER':
                struct_name = self.current.value

                if self.next() and self.current.value == '{':
                    self.parse_struct(begin, struct_name)

                    return None

        while (not self.current.is_punctuation('(')
               or self.next()
               and self.current.is_punctuation('*')):
            if not self.bracket_stack and self.current.is_punctuation(';'):
                return None

            return_type.append(self.current.value)
            self.next()

        if not return_type:
            return None

        func_name = return_type.pop()

        if func_name not in self.functions:
            return None

        while (self.bracket_stack
               or self.next()
               and self.current.is_punctuation('(')):
            self.next()

        code = self.read_source_code(begin, self.previous.span[1]) + ';'

        return func_name, code

    def mark_for_erase(self, begin, end):
        self.chunks_to_erase.append((begin, end))

    def read_source_code(self, begin, end):
        if self.chunks_to_erase:
            chunks = []
            offset = begin

            for chunk_begin, chunk_end in self.chunks_to_erase:
                if chunk_end < offset:
                    continue

                chunks.append(self.source_code[offset:chunk_begin])
                offset = chunk_end

            chunks.append(self.source_code[offset:end])
            self.chunks_to_erase = []
            code = ''.join(chunks)
        else:
            code = self.source_code[begin:end]

        return code.strip()
